from bygg.action import ActionContext, action
from bygg.common_types import CommandStatus
import rich


@action("action2_python")
def example_action(ctx: ActionContext):
    rich.print("[magenta]Running default action from Python.[/magenta]")
    return CommandStatus(0, "All ok.", None)



from typing import Set

import rich
import rich.box
import rich.progress
import rich.status
import rich.table

from bygg.action import Action
from bygg.common_types import CommandStatus
from bygg.runner import JobStatus

console = rich.console.Console()

progress: rich.progress.Progress = rich.progress.Progress(
    rich.progress.SpinnerColumn(
        table_column=rich.table.Column(ratio=None),
        spinner_name="bouncingBar",
        style="blue",
    ),
    rich.progress.TimeElapsedColumn(table_column=rich.table.Column(ratio=None)),
    rich.progress.TextColumn(
        "[cyan]{task.description}",
        table_column=rich.table.Column(ratio=1),
    ),
    transient=True,
    console=console,
)

task_id = None


def on_runner_status(message: str):
    console.print(f"[blue]{message}")


running_jobs: Set[str] = set()
processed_jobs = 0


def print_job_ended(
    name: str, job_status: JobStatus, action: Action, status: CommandStatus | None
):
    job_prefix = "[red]✗[/red]" if job_status == "failed" else "[green]✓[/green]"
    status_color = "red" if job_status == "failed" else "green"

    status_message = status.message if status and status.message else ""

    table = rich.table.Table(show_header=False, pad_edge=False, box=None)
    table.add_column()
    table.add_row(
        f"{job_prefix} [{status_color}]{name}[/{status_color}]"
        f"{': [cyan]' + status_message + '[/cyan]' }"
        f"{' (start) ---' if status and status.output else ''}"
    )
    if status and status.output:
        table.add_row(status.output)
        table.add_row(f"--- [{status_color}]{name}[/{status_color}] (end) ---")
        table.add_row()

    console.print(table)


def on_job_status(
    name: str, job_status: JobStatus, action: Action, status: CommandStatus | None
):
    global processed_jobs
    global task_id
    if task_id is None:
        task_id = progress.add_task("[blue]Building[/blue]")

    if job_status == "failed":
        processed_jobs += 1
        if name in running_jobs:
            running_jobs.remove(name)
        print_job_ended(name, job_status, action, status)
    elif job_status == "running":
        running_jobs.add(name)
        progress.update(
            task_id,
            description=f"[blue]Finished [yellow]{processed_jobs}[/yellow] jobs.[/blue]",
        )
    else:
        processed_jobs += 1
        print_job_ended(name, job_status, action, status)
        if name in running_jobs:
            running_jobs.remove(name)



from bygg.action import ActionContext, action
from bygg.common_types import CommandStatus


@action("testfile 2", message="testfile 2", outputs=["output2.txt"])
def write_foo_to_file(ctx: ActionContext):
    for filename in ctx.outputs:
        with open(filename, "w") as f:
            f.write("foo")
    return CommandStatus(0, "Wrote foo to file.", None)


@action("foo", message="foo", outputs={"foo.txt"})
def foo(ctx: ActionContext):
    print("foo called")
    print(ctx.message)
    return CommandStatus(0, "foo", None)



from pathlib import Path
from typing import Dict, List, Literal, Set

import msgspec

from bygg.action import Action, CommandStatus
from bygg.cache import Cache
from bygg.dag import Dag, create_dag
from bygg.digest import calculate_dependency_digest, calculate_digest


class Job:
    name: str
    action: Action
    status: CommandStatus | None

    def __init__(self, action: Action):
        self.name = action.name
        self.action = action
        self.status = None

    def __repr__(self) -> str:
        return f'"{self.name}, status: {self.status.rc if self.status else "unknown"}"'

    def __hash__(self) -> int:
        return hash(self.name)

    def __str__(self) -> str:
        return self.name

    def __eq__(self, __o: object) -> bool:
        return isinstance(__o, Job) and self.name == __o.name


class ActionDigestItem(msgspec.Struct):
    inputs_digest: str
    outputs_digest: str


class Scheduler:
    cache: Cache
    build_actions: Dict[str, Action]

    job_graph: Dag
    dirty_jobs: Set[str]
    ready_jobs: Set[str]
    running_jobs: Dict[str, Job]
    finished_jobs: Dict[str, Job]

    started: bool
    always_make: bool

    def __init__(self):
        self.cache = Cache()
        self.build_actions = {}
        self.job_graph = create_dag()
        self.dirty_jobs = set()
        self.ready_jobs = set()
        self.running_jobs = {}
        self.finished_jobs = {}
        self.started = False
        self.always_make = False

    def init_cache(self, cache_file: Path):
        self.cache = Cache(cache_file)

    def prepare_run(self, entrypoint: str):
        self.job_graph.clear()
        self.ready_jobs = set()
        self.running_jobs = {}
        self.finished_jobs = {}

        self.job_graph.build_action_graph(
            self.build_actions, self.build_actions[entrypoint]
        )

        # Fill the actions' _dependency_files from their dependencies
        for action in self.build_actions.values():
            for dependency in action.dependencies:
                action.dependency_files.update(self.build_actions[dependency].outputs)
            # print(f"Action {action.name} inputs: {action._dependency_files}")

    def start_run(self, entrypoint: str, *, always_make: bool = False):
        self.always_make = always_make
        self.prepare_run(entrypoint)
        self.cache.load()
        self.started = True

    def shutdown(self):
        self.cache.save()

    def run_status(self) -> Literal["not started", "running", "finished", "failed"]:
        """Check the status of the run"""
        if not self.started:
            return "not started"
        if (
            len(
                [
                    x
                    for x in self.finished_jobs.values()
                    if x.status is not None and x.status.rc != 0
                ]
            )
            > 0
        ):
            return "failed"
        if len(self.job_graph) == 0:
            return "finished"
        return "running"

    def check_dirty(self, job_name: str) -> bool:
        """Check if a job needs to be built."""
        if self.always_make:
            return True

        action = self.build_actions[job_name]
        cached_digests = self.cache.get_digests(job_name)

        if (
            len(action.inputs) == 0
            and len(action.outputs) == 0
            and not action.dynamic_dependency
        ):
            # An action with neither inputs nor outputs will always be built
            # print(f"Job {job_name} is dirty (no inputs or outputs)")
            return True

        if (
            not cached_digests
            or not cached_digests.outputs_digest
            or not cached_digests.inputs_digest
        ):
            # No previous result, so we need to build
            # print(f"Job {job_name} is dirty (no previous result)")
            return True

        outputs_digest, files_were_missing = calculate_dependency_digest(action.outputs)
        if files_were_missing or cached_digests.outputs_digest != outputs_digest:
            # The output has changed, so we need to rebuild
            # print(f"Job {job_name} is dirty (output changed)")
            return True

        inputs_digest, files_were_missing = calculate_dependency_digest(
            action.dependency_files
        )

        # Calculate dynamic dependencies
        if action.dynamic_dependency:
            if not cached_digests.dynamic_digest:
                return True

            dd_result = action.dynamic_dependency()
            if (
                dd_result is None
                or calculate_digest([dd_result]) != cached_digests.dynamic_digest
            ):
                # print(f"Job {job_name} is dirty (dynamic dependency changed)")
                return True

        # TODO handle files_were_missing here? abort?
        if cached_digests.inputs_digest != inputs_digest:
            # The inputs have changed, so we need to rebuild
            # print(f"Job {job_name} is dirty (inputs changed)")
            return True

        # print(f"Job {job_name} is clean")
        return False

    def get_ready_jobs(self, batch_size: int = 0) -> List[Job]:
        """
        Create a batch of jobs and put them in the running pool. Returns all ready jobs
        if batch_size is 0.

        An empty job list may be returned even if there are more jobs left to run if all
        jobs in the batch were skipped. Job runners should continue polling for more
        jobs until the scheduler status is "finished".
        """
        if batch_size == 0 or len(self.ready_jobs) < batch_size:
            new_jobs = self.job_graph.get_ready_jobs(
                self.finished_jobs, self.running_jobs
            )
            dirty_jobs = []
            for job in new_jobs:
                if self.check_dirty(job):
                    dirty_jobs.append(job)
                else:
                    self.skip_job(job)
            self.ready_jobs.update(dirty_jobs)

        if len(self.ready_jobs) == 0:
            return []

        job_list = []
        for _ in range(
            len(self.ready_jobs)
            if batch_size == 0
            else min(batch_size, len(self.ready_jobs))
        ):
            action = self.build_actions[self.ready_jobs.pop()]
            new_job = Job(action)
            self.running_jobs[new_job.name] = new_job
            job_list.append(new_job)

        return job_list

    def skip_job(self, job_name: str):
        """Skip a job, remove it from the graph and don't send it to the runner."""
        self.job_graph.remove_node(job_name)

    def job_finished(self, job: Job):
        """Move a job from the running pool to the finished pool"""
        self.running_jobs.pop(job.name)
        self.finished_jobs[job.name] = job

        if job.status and job.status.rc == 0:
            self.job_graph.remove_node(job.name)
            inputs_digest, _ = calculate_dependency_digest(job.action.dependency_files)
            outputs_digest, _ = calculate_dependency_digest(job.action.outputs)
            dynamic_digest = (
                calculate_digest([dd])
                if job.action.dynamic_dependency
                and (dd := job.action.dynamic_dependency())
                else None
            )

            self.cache.set_digests(
                job.name, inputs_digest, outputs_digest, dynamic_digest
            )
        else:
            self.cache.remove_digests(job.name)


scheduler = Scheduler()



from bygg.action import Action
from bygg.util import create_shell_command, filenames_from_pattern


def test_filenames_from_pattern_1():
    input_files = ["/foo/bar/baz.txt.j2"]
    in_pattern = "%.j2"
    out_pattern = "out_%"
    output_files = [("/foo/bar/baz.txt.j2", "/foo/bar/out_baz.txt")]
    assert filenames_from_pattern(input_files, in_pattern, out_pattern) == output_files


def test_filenames_from_pattern_2():
    input_files = ["/foo/bar/baz.txt.j2"]
    in_pattern = "%.txt.j2"
    out_pattern = "out_%.txt"
    output_files = [("/foo/bar/baz.txt.j2", "/foo/bar/out_baz.txt")]
    assert filenames_from_pattern(input_files, in_pattern, out_pattern) == output_files


def test_filenames_from_pattern_3():
    input_files = ["foo.c", "bar.c"]
    in_pattern = "%.c"
    out_pattern = "%.o"
    output_files = [("foo.c", "foo.o"), ("bar.c", "bar.o")]
    assert filenames_from_pattern(input_files, in_pattern, out_pattern) == output_files


def test_filenames_from_pattern_4():
    input_files = ["foo.txt", "bar.txt"]
    in_pattern = "%"
    out_pattern = "out_%"
    output_files = [("foo.txt", "out_foo.txt"), ("bar.txt", "out_bar.txt")]
    assert filenames_from_pattern(input_files, in_pattern, out_pattern) == output_files


def test_shell_command():
    shell_command = create_shell_command("echo 'shell action output'")

    status = shell_command(Action("test context", inputs=set(), outputs=set()))
    assert status.rc == 0
    assert status.output == "shell action output\n"



from dataclasses import dataclass
from pathlib import Path
import pickle

from bygg.scaffolding import STATUS_DIR, make_sure_status_dir_exists

DEFAULT_DB_FILE = STATUS_DIR / "cache.db"


@dataclass
class InputsOutputsDigests:
    inputs_digest: str
    outputs_digest: str
    dynamic_digest: str | None


@dataclass
class CacheState:
    digests: dict[str, InputsOutputsDigests]


class Cache:
    data: CacheState | None
    db_file: Path

    def __init__(self, db_file: Path = DEFAULT_DB_FILE):
        self.data = CacheState({})
        self.db_file = db_file

    def load(self):
        make_sure_status_dir_exists()
        try:
            with open(self.db_file, "rb") as f:
                self.data = pickle.load(f)
        except (EOFError, FileNotFoundError):
            self.data = CacheState({})

    def save(self):
        if not self.data:
            return
        with open(self.db_file, "wb") as f:
            pickle.dump(self.data, f)
        # print(f"Cache: {self.data.digests}")

    def get_digests(self, name: str) -> InputsOutputsDigests | None:
        if not self.data:
            return None
        return self.data.digests.get(name, None)

    def set_digests(
        self,
        name: str,
        inputs_digest: str,
        outputs_digest: str,
        dynamic_digest: str | None = None,
    ):
        if not self.data:
            return
        self.data.digests[name] = InputsOutputsDigests(
            inputs_digest, outputs_digest, dynamic_digest
        )

    def remove_digests(self, name: str):
        if not self.data:
            return
        self.data.digests.pop(name, None)



import os
from pathlib import Path
from tempfile import mkstemp

from bygg.action import Action, CommandStatus
from bygg.scheduler import scheduler


def get_closed_tmpfile() -> Path:
    fd, path = mkstemp()
    os.close(fd)
    return Path(path)


def test_scheduler_single_action():
    scheduler.__init__()
    scheduler.init_cache(get_closed_tmpfile())

    Action(
        name="action1",
        is_entrypoint=True,
    )

    assert scheduler.run_status() == "not started"

    scheduler.start_run("action1")

    assert scheduler.run_status() == "running"

    # check that the scheduler has the correct number of actions
    assert len(scheduler.build_actions) == 1
    assert len(scheduler.job_graph) == 1

    job = scheduler.get_ready_jobs(1)[0]
    assert job
    assert job.action.name == "action1"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 0
    assert scheduler.run_status() == "finished"


def test_scheduler_nonbranching():
    scheduler.__init__()
    scheduler.init_cache(get_closed_tmpfile())

    Action(
        name="action1",
        dependencies=["action2"],
        is_entrypoint=True,
    )
    Action(
        name="action2",
        dependencies=["action3"],
    )
    Action(
        name="action3",
        dependencies=["action4"],
    )
    Action(
        name="action4",
        dependencies=[],
    )

    assert scheduler.run_status() == "not started"

    scheduler.start_run("action1")

    assert scheduler.run_status() == "running"

    # check that the scheduler has the correct number of actions
    assert len(scheduler.build_actions) == 4
    assert len(scheduler.job_graph) == 4

    job = scheduler.get_ready_jobs(1)[0]
    assert job
    assert job.action.name == "action4"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 3
    assert scheduler.run_status() == "running"

    job = scheduler.get_ready_jobs(1)[0]
    assert job
    assert job.action.name == "action3"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 2
    assert scheduler.run_status() == "running"

    job = scheduler.get_ready_jobs(1)[0]
    assert job
    assert job.action.name == "action2"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 1
    assert scheduler.run_status() == "running"

    job = scheduler.get_ready_jobs(1)[0]
    assert job
    assert job.action.name == "action1"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 0
    assert scheduler.run_status() == "finished"

    assert len(scheduler.build_actions) == 4
    assert len(scheduler.running_jobs) == 0
    assert len(scheduler.finished_jobs) == 4


def test_scheduler_branching():
    scheduler.__init__()
    scheduler.init_cache(get_closed_tmpfile())

    Action(
        name="action1",
        dependencies=["action2", "action3"],
        is_entrypoint=True,
    )
    Action(
        name="action2",
        dependencies=["action4"],
    )
    Action(
        name="action3",
        dependencies=["action4"],
    )
    Action(
        name="action4",
        dependencies=[],
    )

    assert scheduler.run_status() == "not started"

    scheduler.start_run("action1")

    assert scheduler.run_status() == "running"

    # check that the scheduler has the correct number of actions
    assert len(scheduler.build_actions) == 4
    assert len(scheduler.job_graph) == 4

    jobs = scheduler.get_ready_jobs(1)
    assert jobs
    job = jobs[0]
    assert job.action.name == "action4"
    assert not scheduler.get_ready_jobs(1)
    assert scheduler.run_status() == "running"

    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 3
    assert scheduler.run_status() == "running"

    job1 = scheduler.get_ready_jobs(1)[0]
    assert job1 is not None
    assert job1.action.name == "action2" or job1.action.name == "action3"
    job1.status = CommandStatus(0, "Executed successfully", None)
    assert scheduler.run_status() == "running"

    job2 = scheduler.get_ready_jobs(1)[0]
    assert job2 is not None
    assert job2.action.name == "action2" or job2.action.name == "action3"
    job2.status = CommandStatus(0, "Executed successfully", None)
    assert scheduler.run_status() == "running"

    assert job1.name != job2.name

    assert not scheduler.get_ready_jobs(1)
    scheduler.job_finished(job1)
    scheduler.job_finished(job2)
    assert len(scheduler.job_graph) == 1
    assert scheduler.run_status() == "running"

    job = scheduler.get_ready_jobs(1)[0]
    assert job is not None
    assert job.action.name == "action1"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 0
    assert scheduler.run_status() == "finished"

    assert len(scheduler.build_actions) == 4
    assert len(scheduler.running_jobs) == 0
    assert len(scheduler.finished_jobs) == 4


def test_scheduler_branching_one_failed():
    scheduler.__init__()
    scheduler.init_cache(get_closed_tmpfile())

    Action(
        name="action1",
        dependencies=["action2", "action3"],
        is_entrypoint=True,
    )
    Action(
        name="action2",
        dependencies=["action4"],
    )
    Action(
        name="action3",
        dependencies=["action4"],
    )
    Action(
        name="action4",
        dependencies=[],
    )

    assert scheduler.run_status() == "not started"

    scheduler.start_run("action1")

    assert scheduler.run_status() == "running"

    # check that the scheduler has the correct number of actions
    assert len(scheduler.build_actions) == 4
    assert len(scheduler.job_graph) == 4

    job = scheduler.get_ready_jobs(1)[0]
    assert job is not None
    assert job.action.name == "action4"
    assert not scheduler.get_ready_jobs(1)
    assert scheduler.run_status() == "running"

    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 3

    job1 = scheduler.get_ready_jobs(1)[0]
    assert job1 is not None
    assert job1.action.name == "action2" or job1.action.name == "action3"
    job1.status = CommandStatus(0, "Executed successfully", None)
    assert scheduler.run_status() == "running"

    job2 = scheduler.get_ready_jobs(1)[0]
    assert job2 is not None
    assert job2.action.name == "action2" or job2.action.name == "action3"
    job2.status = CommandStatus(1, "Failed", None)
    assert scheduler.run_status() == "running"

    assert job1.name != job2.name

    assert not scheduler.get_ready_jobs(1)
    scheduler.job_finished(job1)
    scheduler.job_finished(job2)
    assert scheduler.run_status() == "failed"

    assert not scheduler.get_ready_jobs(1)
    assert len(scheduler.ready_jobs) == 0
    assert len(scheduler.job_graph) == 2
    assert scheduler.run_status() == "failed"

    assert len(scheduler.build_actions) == 4
    assert len(scheduler.running_jobs) == 0
    assert len(scheduler.finished_jobs) == 3


def test_scheduler_dynamic_dependency():
    scheduler.__init__()
    cache_file = get_closed_tmpfile()
    scheduler.init_cache(cache_file)

    Action(
        name="action1",
        dependencies=["action2"],
        is_entrypoint=True,
    )
    Action(name="action2", dynamic_dependency=lambda: "foo")

    assert scheduler.run_status() == "not started"
    scheduler.start_run("action1")
    assert scheduler.run_status() == "running"

    # check that the scheduler has the correct number of actions
    assert len(scheduler.build_actions) == 2
    assert len(scheduler.job_graph) == 2

    job = scheduler.get_ready_jobs()[0]
    assert job
    assert job.name == "action2"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)

    # check again that the scheduler has the correct number of actions
    assert len(scheduler.build_actions) == 2
    assert len(scheduler.job_graph) == 1

    job = scheduler.get_ready_jobs()[0]
    assert job
    assert len(scheduler.job_graph) == 1
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)

    assert scheduler.run_status() == "finished"

    assert len(scheduler.build_actions) == 2
    assert len(scheduler.running_jobs) == 0


def test_scheduler_dynamic_dependency_two_runs():
    scheduler.__init__()
    cache_file = get_closed_tmpfile()
    scheduler.init_cache(cache_file)

    Action(
        name="action1",
        dependencies=["action2"],
        is_entrypoint=True,
    )
    Action(name="action2", dynamic_dependency=lambda: "foo")

    assert scheduler.run_status() == "not started"

    scheduler.start_run("action1")

    assert scheduler.run_status() == "running"

    # check that the scheduler has the correct number of actions
    assert len(scheduler.build_actions) == 2
    assert len(scheduler.job_graph) == 2

    job = scheduler.get_ready_jobs(1)[0]
    assert job
    assert job.action.name == "action2"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 1
    assert scheduler.run_status() == "running"

    job = scheduler.get_ready_jobs(1)[0]
    assert job
    assert job.action.name == "action1"
    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)
    assert len(scheduler.job_graph) == 0
    assert scheduler.run_status() == "finished"

    assert len(scheduler.build_actions) == 2
    assert len(scheduler.running_jobs) == 0
    assert len(scheduler.finished_jobs) == 2

    scheduler.shutdown()

    # Second run

    scheduler.__init__()
    scheduler.init_cache(cache_file)

    Action(
        name="action1",
        dependencies=["action2"],
        is_entrypoint=True,
    )
    Action(name="action2", dynamic_dependency=lambda: "foo")

    assert scheduler.run_status() == "not started"

    scheduler.start_run("action1")

    assert scheduler.run_status() == "running"

    # check that the scheduler has the correct number of actions
    assert len(scheduler.build_actions) == 2
    assert len(scheduler.job_graph) == 2

    jobs = scheduler.get_ready_jobs()
    assert len(jobs) == 0
    assert scheduler.run_status() == "running"

    # Ask again since the job should have been skipped and the scheduler is still
    # running:
    job = scheduler.get_ready_jobs()[0]
    assert job
    assert len(scheduler.job_graph) == 1

    job.status = CommandStatus(0, "Executed successfully", None)
    scheduler.job_finished(job)

    assert scheduler.run_status() == "finished"

    assert len(scheduler.build_actions) == 2
    assert len(scheduler.running_jobs) == 0



from collections import deque
from typing import Any, Dict, Iterable, List, Set

from bygg.action import Action


class Dag:
    def __len__(self):
        return 0

    def clear(self):
        ...

    def remove_node(self, node: str):
        ...

    def build_action_graph(self, build_actions: Dict[str, Action], action: Action):
        ...

    def get_ready_jobs(
        self,
        finished_jobs: Dict[str, Any],
        running_jobs: Dict[str, Any],
    ) -> List[str]:
        ...

    def get_all_jobs(self) -> Iterable[str]:
        ...


class GraphlibDag(Dag):
    nodes: Set[str]

    def __init__(self):
        import graphlib

        self.graph = graphlib.TopologicalSorter()
        self.nodes = set()

    def __len__(self):
        return len(self.nodes)

    def clear(self):
        import graphlib

        self.graph = graphlib.TopologicalSorter()
        self.nodes = set()

    def remove_node(self, node: str):
        self.graph.done(node)
        self.nodes.remove(node)

    def build_action_graph(self, build_actions: Dict[str, Action], action: Action):
        """Build the action graph."""
        queue = deque([action])
        while len(queue) > 0:
            a = queue.popleft()
            self.graph.add(a.name)
            self.nodes.add(a.name)
            for dependency in a.dependencies:
                dependency_action = build_actions.get(dependency)
                if not dependency_action:
                    raise ValueError(f"Action '{dependency}' not found")
                queue.append(dependency_action)
                self.graph.add(a.name, dependency)
        self.graph.prepare()

    def get_ready_jobs(
        self,
        finished_jobs: Dict[str, Any],
        running_jobs: Dict[str, Any],
    ) -> List[str]:
        return list(self.graph.get_ready())

    def get_all_jobs(self) -> Iterable[str]:
        return list(self.nodes)


def create_dag() -> Dag:
    return GraphlibDag()



import os
import signal
from typing import Callable, List, Literal

import multiprocess.managers  # type: ignore
from multiprocess.pool import ApplyResult, Pool  # type: ignore

from bygg.action import Action, CommandStatus
from bygg.scheduler import Job, scheduler

JobStatus = Literal["queued", "running", "finished", "failed", "stopped", "skipped"]
JobStatusListener = Callable[[str, JobStatus, Action, CommandStatus | None], None]
RunnerStatusListener = Callable[[str], None]


class ProcessRunner:
    job_status_listener: JobStatusListener
    runner_status_listener: RunnerStatusListener

    def __init__(self):
        self.job_status_listener = lambda *args: None
        self.runner_status_listener = lambda *args: None

    def start(self, max_workers: int = 1) -> bool:
        self.runner_status_listener(
            f"[blue]Starting process runner with {max_workers} threads"
        )

        def init_worker():
            """Ignore CTRL+C in the worker process."""
            signal.signal(signal.SIGINT, signal.SIG_IGN)

        with Pool(max_workers, init_worker) as pool:
            scheduled_queue: List[ApplyResult] = []
            backlog: List[Job] = []

            manager = multiprocess.managers.SyncManager()
            manager.start()
            qq = manager.Queue()  # type: ignore # no/bad multiprocess types

            def poll_msg_queue():
                while not qq.empty():
                    msg, job_msg = qq.get_nowait()
                    self.job_status_listener(
                        job_msg.name, "running", job_msg.action, None
                    )

            while True:
                if len(backlog) < 2 * max_workers and (
                    jobs := scheduler.get_ready_jobs()
                ):
                    backlog += jobs

                if (
                    len(scheduled_queue) == 0
                    and len(backlog) == 0
                    and scheduler.run_status() == "finished"
                ):
                    return True

                # Keep the scheduled queue relatively short since we're polling
                batch_size = max_workers * 2
                if len(scheduled_queue) < batch_size:
                    for job in backlog[:batch_size]:
                        # Skip jobs with no command or ones that are clean
                        if job.action.command is None:
                            job.status = CommandStatus(
                                0,
                                f"{'No command, skipping'}",
                                None,
                            )
                            scheduler.job_finished(job)
                            self.job_status_listener(
                                job.name, "skipped", job.action, job.status
                            )
                            backlog.remove(job)
                            continue

                        # Run in-process
                        if job.action.scheduling_type == "in-process":
                            self.job_status_listener(
                                job.name, "running", job.action, None
                            )
                            job.status = job.action.command(job.action)
                            scheduler.job_finished(job)
                            self.job_status_listener(
                                job.name,
                                "finished" if job.status.rc == 0 else "failed",
                                job.action,
                                job.status,
                            )
                            backlog.remove(job)
                            continue

                        # Schedule job to be run on the worker processes
                        backlog.remove(job)
                        scheduled_queue.append(pool.apply_async(run_job, (job, qq)))

                poll_msg_queue()

                if len(scheduled_queue) == 0:
                    continue

                queued_job = scheduled_queue.pop(0)
                queued_job.wait(0.1)
                if not queued_job.ready():
                    scheduled_queue.append(queued_job)  # add it back to the queue
                else:
                    job_result = queued_job.get()
                    missing_files = self.check_for_missing_output_files(job_result)
                    if missing_files:
                        # TODO This should be a job status, not a runner status
                        self.runner_status_listener(
                            f"[red]Job [bold]{job_result.name}[/bold] didn't create the output file{'s' if len(missing_files) > 1 else ''} that it declared: [bold]{', '.join(missing_files)}"
                        )

                    scheduler.job_finished(job_result)
                    if job_result.status.rc == 0:
                        self.job_status_listener(
                            job_result.name,
                            "finished",
                            job_result.action,
                            job_result.status,
                        )
                    else:
                        self.job_status_listener(
                            job_result.name,
                            "failed",
                            job_result.action,
                            job_result.status,
                        )
                        self.runner_status_listener("Aborting.")
                        poll_msg_queue()
                        return False

    def check_for_missing_output_files(self, job: Job) -> List[str]:
        missing_files: List[str] = []
        for filename in job.action.outputs:
            if not os.path.exists(filename):
                missing_files.append(filename)
        return missing_files


def run_job(job: Job, qq):
    qq.put_nowait(("START", job))
    try:
        if job.action.command is None:
            job.status = CommandStatus(0, "No command, skipping", None)
        else:
            job.status = job.action.command(job.action)
    except Exception as e:
        job.status = CommandStatus(1, "Job failed with exception.", f"{e}")
    return job


# runner = SingleThreadedRunner()
runner = ProcessRunner()



from dataclasses import dataclass
from typing import Callable, Iterable, Literal, Optional, Set, Tuple

from bygg.common_types import CommandStatus

SchedulingType = Literal["in-process", "processpool"]

# Function that returns a string that is included in the dependency digest. Returning
# None causes the value to be ignored.
DynamicDependency = Callable[[], str | None]


@dataclass
class ActionContext:
    name: str
    message: str | None
    inputs: Set[str]
    outputs: Set[str]
    dependencies: Set[str]
    dynamic_dependency: Optional[DynamicDependency]
    is_entrypoint: bool
    scheduling_type: SchedulingType


Command = Callable[
    [ActionContext],
    CommandStatus,
]


class Action(ActionContext):
    """
    An action in the build graph.

    name: The name of the action.
    message (optional): A message to display when the action is run.
    inputs (optional): A list of paths that are inputs to the action.
    outputs (optional): A list of paths that are outputs of the action.
    dependencies (optional): A list of action names that this action depends on.
    is_entrypoint (optional): Whether this action is an entrypoint to the build graph.
    command (optional): Instance of one of the Command types to run.

    """

    command: Command | None
    dependency_files: Set[str]

    def __init__(
        self,
        name: str,
        message: str | None = None,
        inputs: Optional[Iterable[str]] = None,
        outputs: Optional[Iterable[str]] = None,
        dependencies: Optional[Iterable[str]] = None,
        dynamic_dependency: Optional[DynamicDependency] = None,
        is_entrypoint: bool = False,
        command: Command | None = None,
        scheduling_type: SchedulingType = "processpool",
    ):
        from bygg.scheduler import scheduler

        self.name = name
        self.message = message
        self.inputs = {*inputs} if inputs else set()
        self.outputs = {*outputs} if outputs else set()
        self.dependencies = {*dependencies} if dependencies else set()
        self.dynamic_dependency = dynamic_dependency
        self.is_entrypoint = is_entrypoint
        self.command = command
        self.scheduling_type = scheduling_type

        self.dependency_files = set()

        scheduler.build_actions[name] = self

    def __repr__(self):
        return f"""Action(name={self.name}, inputs=({
                ', '.join([str(i) for i in self.inputs])}), outputs=({
                ', '.join([str(o) for o in self.outputs])}), dependencies=({
                ', '.join(self.dependencies)}) is_entrypoint={
            self.is_entrypoint})"""

    def __str__(self):
        return self.__repr__()


def action(
    name: str,
    *,
    message: Optional[str] = None,
    inputs: Optional[Iterable[str]] = None,
    outputs: Optional[Iterable[str]] = None,
    dependencies: Optional[Iterable[str]] = None,
    dynamic_dependency: Optional[DynamicDependency] = None,
    is_entrypoint: bool = False,
):
    """Decorator for creating an Action from a function."""

    def create_action(func: Callable):
        Action(
            name,
            message,
            inputs=inputs,
            outputs=outputs,
            dependencies=dependencies,
            dynamic_dependency=dynamic_dependency,
            is_entrypoint=is_entrypoint,
            command=func,
        )

    return create_action


def action_set(
    base_name: str,
    *,
    message: Optional[str] = None,
    file_pairs: Iterable[Tuple[str, str]],
    dependencies: Optional[Iterable[str]] = None,
    is_entrypoint: bool = False,
):
    """
    Decorator for creating individual Actions from a list of input-output file pairs and
    a function. Creates one Action that depends on all of the individual Actions. It has
    name = base_name, which in turn can be used as a dependency for other Actions.
    """

    def create_actions(func: Callable):
        action_list = []
        for input_file, output_file in file_pairs:
            action_name = f"{base_name}_{output_file}"
            action_list.append(action_name)
            Action(
                action_name,
                message,
                inputs=[input_file],
                outputs=[output_file],
                dependencies=dependencies,
                is_entrypoint=is_entrypoint,
                command=func,
            )
        Action(base_name, f"Action set {base_name}", dependencies=action_list)

    return create_actions



import os
import sys
from typing import Dict, List, Optional

import msgspec
import rich
import rich.status

PYTHON_INPUTFILE = "Byggfile.py"
YAML_INPUTFILE = "Byggfile.yml"


class SettingsSection(msgspec.Struct, forbid_unknown_fields=True):
    default_action: Optional[str] = None


class ActionItem(msgspec.Struct, forbid_unknown_fields=True):
    """
    name: The name of the action.
    description: A description of the action. Used in e.g. help messages.
    message: A message to print when the action is executed.
    inputs: A list of files that are used as input to the action.
    outputs: A list of files that are generated by the action.
    dependencies: A list of actions that must be executed before this action.
    is_entrypoint: Whether this action is an entrypoint. Entry points are actions that
    can be executed directly from the command line.
    shell: The shell command to execute when the action is run.
    """

    name: str
    description: Optional[str] = None
    message: Optional[str] = None
    inputs: Optional[List[str]] = None
    outputs: Optional[List[str]] = None
    dependencies: Optional[List[str]] = None
    is_entrypoint: Optional[bool] = None
    environment: Optional[str] = "default"
    shell: Optional[str] = None


class Environment(msgspec.Struct, forbid_unknown_fields=True):
    """
    name: The name of the environment.

    description: A description of the environment. Used in e.g. help messages.

    byggfiles: A list of Byggfiles that use this environment.

    inputs: A list of files that are used as input to the environment. Typically pip
    requirements files, but can be any files.

    venv_directory: The directory where the virtual environment is located. Will be
    recreated by Bygg if any of the inputs are modified.

    shell: The shell command for creating the environment.

    entrypoints: A list of entrypoints that use this environment.
    """

    byggfile: str
    inputs: List[str]
    venv_directory: str
    shell: str
    description: Optional[str] = None
    message: Optional[str] = None


class ByggFile(msgspec.Struct, forbid_unknown_fields=True):
    actions: List[ActionItem]
    settings: SettingsSection = msgspec.field(default_factory=SettingsSection)
    environments: Dict[str, Environment] = msgspec.field(default_factory=dict)


def load_python_build_file(build_file: str):
    if os.path.isfile(build_file):
        with open(build_file, "r") as f:
            # modify load path to make the current directory importable
            preamble = "import sys\nsys.path.insert(0, '.')\n\n"
            exec(preamble + f.read(), globals())


def read_config_file() -> ByggFile | None:
    if not os.path.isfile(YAML_INPUTFILE):
        return None

    try:
        with open(YAML_INPUTFILE, "r") as f:
            return msgspec.yaml.decode(f.read(), type=ByggFile)
    except Exception as e:
        rich.print(
            "[red bold]Error while reading configuration file "
            f"[yellow]{YAML_INPUTFILE}[/yellow]:[/red bold] {e}"
        )
        sys.exit(1)


def dump_schema():
    import json

    schema = msgspec.json.schema(ByggFile)
    print(json.dumps(schema, indent=2))



import os
import re
import subprocess
from typing import List, Optional, Set, Tuple

from bygg.common_types import CommandStatus


def create_shell_command(shell_command: str, message: Optional[str] = None):
    def call_shell_command(inputs: Optional[Set[str]], outputs: Optional[Set[str]]):
        # Map stderr onto stdout and return both as the output
        process = subprocess.run(
            shell_command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            encoding="utf-8",
        )
        return CommandStatus(process.returncode, message, process.stdout)

    return lambda ctx: call_shell_command(ctx.inputs, ctx.outputs)


def filenames_from_pattern(
    input_files: List[str], in_pattern: str, out_pattern: str
) -> List[Tuple[str, str]]:
    """
    Generate a list of tuples with the input and output file names from a a list of
    input file names and a pattern like in Makefile pattern rules. File names that don't
    match the pattern are ignored. Example:

      ["/foo/bar/baz.txt.j2"], "%.txt.j2", "out_%.txt")
        => [("/foo/bar/baz.txt.j2", "/foo/bar/out_baz.txt")]
    """

    output_files = []
    matcher = re.compile(in_pattern.replace("%", "(.+)"))
    for input_file in input_files:
        head, tail = os.path.split(input_file)
        matches = matcher.match(tail)
        if matches:
            output_file = os.path.join(head, out_pattern.replace("%", matches.group(1)))
            output_files.append((input_file, output_file))
    return output_files



import argparse
from dataclasses import dataclass
from multiprocessing import cpu_count
import os
import shutil
import stat
import subprocess
import sys
import time
from typing import List

import rich
import rich.status

from bygg.apply_configuration import apply_configuration
from bygg.configuration import (
    PYTHON_INPUTFILE,
    YAML_INPUTFILE,
    ByggFile,
    dump_schema,
    read_config_file,
)
from bygg.runner import runner
from bygg.scheduler import scheduler
from bygg.status_display import on_job_status, on_runner_status, progress


def get_job_count_limit():
    # TODO use affinity instead (on U**X): https://stackoverflow.com/a/55423170
    return cpu_count()


def init_status_listeners():
    """
    Set up status listeners.
    """
    runner.job_status_listener = on_job_status
    runner.runner_status_listener = on_runner_status


def build(actions: List[str], job_count: int | None, always_make: bool) -> bool:
    try:
        init_status_listeners()

        max_workers = get_job_count_limit() if job_count is None else job_count

        for action in actions:
            t1 = time.time()
            rich.print(f"Building action '{action}':")

            progress.start()
            scheduler.start_run(action, always_make=always_make)
            status = runner.start(max_workers)
            progress.disable
            progress.stop()
            scheduler.shutdown()

            if status:
                rich.print(
                    f"[green]Action '{action}' completed in {time.time() - t1:.2f} s."
                )
            else:
                rich.print(
                    f"[red]Action '{action}' failed after {time.time() - t1:.2f} s."
                )

            # cs = build(action)
            # if cs is None:
            #     rich.print(f"Action '{action}' is up to date.")
            # else:
            #     rich.print(f"Action '{action}' completed with return code {cs.rc}.")
            #     rich.print(
            #         f"{len(cs.changed_files)} files changed in {time() - t1:.2f} s."
            #     )
            # rich.print("=========================================")
    except KeyboardInterrupt:
        rich.print("\n[yellow]Build was interrupted by user.[/yellow]")
        return False
    except KeyError as e:
        rich.print(f"[red]Error: Action '{e}' not found.[/red]")
        return False
    finally:
        progress.stop()
        scheduler.shutdown()

    return True


def clean(actions: List[str]):
    try:
        init_status_listeners()

        for action in actions:
            rich.print(f"Cleaning action '{action}':")
            scheduler.prepare_run(action)
            for job_name in scheduler.job_graph.get_all_jobs():
                job = scheduler.build_actions.get(job_name, None)
                if job is None:
                    continue
                for output in job.outputs:
                    try:
                        s = os.stat(output)
                        if stat.S_ISREG(s.st_mode):
                            os.unlink(output)
                        elif stat.S_ISDIR(s.st_mode):
                            rich.print(f"Removing directory: {output}")
                            shutil.rmtree(output)
                    except FileNotFoundError:
                        pass
            scheduler.shutdown()
    except KeyboardInterrupt:
        rich.print("[yellow]Build was interrupted by user.[/yellow]")
        return False
    except KeyError as e:
        rich.print(f"[yellow]Error: Action '{e}' not found.[/yellow]")
        return False
    finally:
        progress.stop()
        scheduler.shutdown()

    return True


def check(actions: List[str]):
    print("check")
    t0 = time.time()
    init_status_listeners()

    for action in actions:
        t1 = time.time()
        with rich.status.Status(f"Preparing action '{action}':"):
            scheduler.prepare_run(action)
        rich.print(f"Action '{action}' prepared in {time.time() - t1:.2f} s.")

    rich.print(f"Total time for --check was {time.time() - t0:.2f} s.")
    return True


def list_actions() -> bool:
    entrypoints = [x for x in scheduler.build_actions.values() if x.is_entrypoint]

    if entrypoints:
        rich.print("Available actions:")
        for action in sorted(entrypoints, key=lambda x: x.name):
            rich.print(f"  {action.name}")
        return True
    else:
        program_name = os.path.basename(sys.argv[0])
        rich.print("[yellow]Loaded build files but no entrypoints were found.[/yellow]")
        rich.print(f"[yellow]Type `{program_name} --help` for help.[/yellow]")
        return False


def print_no_actions_text(configuration: ByggFile | None):
    return list_actions()


def print_version():
    import importlib.metadata

    print(f"{__package__} {importlib.metadata.version(__package__)}")


MAKE_COMPATIBLE_PANEL = "(Roughly) Make-compatible options"


def dispatcher(args: argparse.Namespace):
    """
    A build tool written in Python, where all actions can be written in Python.
    """
    if args.dump_schema:
        dump_schema()
        sys.exit(0)

    if args.version:
        print_version()
        sys.exit(0)

    directory = args.directory[0] if args.directory else None
    is_restarted_with_env = (
        args.is_restarted_with_env[0] if args.is_restarted_with_env else None
    )

    if directory and not is_restarted_with_env:
        directory_arg = args.directory[0]
        rich.print(f"Entering directory '{directory_arg}'")
        os.chdir(directory_arg)

    if not os.path.isfile(PYTHON_INPUTFILE) and not os.path.isfile(YAML_INPUTFILE):
        rich.print("No build files found.")
        sys.exit(1)

    configuration = read_config_file()
    action_partitions = partition_actions(configuration, args.actions)

    if not action_partitions:
        if os.path.isfile(PYTHON_INPUTFILE):
            # No configuration file, so load the Python build file directly.
            apply_configuration(None, None, None)
            status = do_dispatch(args, args.actions)
        else:
            status = print_no_actions_text(configuration)
        if status:
            sys.exit(0)
        sys.exit(1)

    for partition in action_partitions:
        env = partition.environment_name
        # Check if we should restart with another Python interpreter (e.g. from a
        # virtualenv):
        restart_with = apply_configuration(configuration, env, is_restarted_with_env)

        if restart_with is not None and not is_restarted_with_env:
            exec_list = construct_exec_list(args, restart_with, partition)
            try:
                process = subprocess.run(exec_list, encoding="utf-8")
                if process.returncode != 0:
                    sys.exit(process.returncode)
            except FileNotFoundError:
                rich.print(
                    f"[red]Error: Could not restart with '{restart_with}'.[/red]\n"
                    f"[yellow]Please make sure that bygg is in your pip requirements list for this environment.[/yellow]"
                )
                sys.exit(1)
        else:
            status = do_dispatch(args, partition.actions)
            if not status:
                sys.exit(1)


def do_dispatch(args: argparse.Namespace, actions: List[str]) -> bool:
    if args.list or not actions:
        status = list_actions()
    elif args.check:
        status = check(actions)
    elif args.clean:
        status = clean(actions)
    else:
        jobs = int(args.jobs) if args.jobs else None
        status = build(actions, jobs, args.always_make)

    return status


@dataclass
class ActionPartition:
    """
    environment_name: The name of the environment that the actions should be run in.
    None means the implicit default environment, i.e. typically the base process.

    actions: The actions that should be run in the environment.
    """

    environment_name: str | None
    actions: List[str]


def construct_exec_list(
    args: argparse.Namespace, restart_with: str, partition: ActionPartition
):
    exec_list = [restart_with, *partition.actions]
    for k, v in vars(args).items():
        if k == "actions":
            continue
        elif v is True:
            exec_list.append(f"--{k}")
        elif v:
            exec_list.append(f"--{k} {v}")
    if partition.environment_name:
        exec_list.append("--is_restarted_with_env")
        exec_list.append(partition.environment_name)
    return exec_list


def partition_actions(
    configuration: ByggFile | None,
    actions: List[str] | None,
) -> List[ActionPartition] | None:
    """
    Partition the actions into groups that should be run in the same environment.
    """

    # TODO: Implement this.

    resolved_actions = actions if actions else []

    if configuration:
        if not actions and configuration.settings.default_action is not None:
            # Resolve to default action:
            resolved_actions += [configuration.settings.default_action]
            return [ActionPartition("default", resolved_actions)]

        # Put consecutive actions with the same environment in the same partition:
        action_partitions = []
        action_dict = {a.name: a for a in configuration.actions}
        current_partition: ActionPartition | None = None
        for action in [action_dict[a] for a in resolved_actions]:
            if (
                current_partition is None
                or current_partition.environment_name != action.environment
            ):
                current_partition = ActionPartition(action.environment, [action.name])
                action_partitions.append(current_partition)
            else:
                current_partition.actions.append(action.name)

        return action_partitions


def create_argument_parser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""
A build tool written in Python, where all actions can be written in Python.

Build the default action:
 %(prog)s

Build ACTION:
 %(prog)s ACTION

Clean ACTION:
 %(prog)s ACTION --clean

List available actions:
 %(prog)s --list
""",
    )
    parser.add_argument(
        "actions",
        nargs="*",
        default=None,
        help="Entrypoint actions to operate on.",
    )
    parser.add_argument(
        "-v", "--version", action="store_true", help="Show version string and exit."
    )
    parser.add_argument(
        "--is_restarted_with_env",
        nargs=1,
        type=str,
        default=None,
        help=argparse.SUPPRESS,
    )
    # Commands that operate on the build setup:
    build_setup_wrapper_group = parser.add_argument_group(
        "Commands that operate on the build setup"
    )  # add_mutually_exclusive_group doesn't accept a title, so wrap it in a regular group.
    build_setup_group = build_setup_wrapper_group.add_mutually_exclusive_group()
    build_setup_group.add_argument(
        "-c",
        "--check",
        action="store_true",
        help="Check the specified actions.",
    )
    build_setup_group.add_argument(
        "--clean",
        action="store_true",
        help="Clean the outputs of the specified actions.",
    )
    build_setup_group.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="List available actions.",
    )
    # Some arguments inspired by Make:
    make_group = parser.add_argument_group("Make-like arguments")
    make_group.add_argument(
        "-C",
        "--directory",
        nargs=1,
        type=str,
        default=None,
        help="Change to the specified directory.",
    )
    make_group.add_argument(
        "-j",
        "--jobs",
        nargs="?",
        type=int,
        default=None,
        help="Specify the number of jobs to run simultaneously. None means to use the number of available cores.",
    )
    make_group.add_argument(
        "-B",
        "--always-make",
        action="store_true",
        help="Always build all actions.",
    )
    # Meta arguments:
    meta_group = parser.add_argument_group("Meta arguments")
    meta_group.add_argument(
        "--dump-schema",
        action="store_true",
        help="Generate a JSON Schema for the Byggfile.yml files. The schema will be printed to stdout.",
    )
    return parser


def main():
    parser = create_argument_parser()
    args = parser.parse_args()

    try:
        return dispatcher(args)
    except KeyboardInterrupt:
        rich.print("[red]Interrupted by user. Aborting.[/red]")
        return 1


if __name__ == "__main__":
    main()



from dataclasses import dataclass


@dataclass
class CommandStatus:
    """The status of a command."""

    rc: int  # return code; follows shell conventions where 0 is success
    message: str | None  # a message to display to the user
    output: str | None  # output of the command



import nox

nox.options.error_on_external_run = True


@nox.session(python=["3.11"])
def tests(session):
    session.install("-r", "requirements.txt", "-r", "requirements-dev.txt")
    session.install(".")
    session.run("pytest")


@nox.session(python=["3.11"])
def basics(session):
    session.install(".")
    session.run("bygg", success_codes=[1], silent=True)
    session.run("bygg", "--help", silent=True)


@nox.session(python=["3.11"])
def examples(session):
    session.install(".")

    with session.chdir("examples/only_python"):
        session.run("bygg")
        session.run("bygg", "--clean")

    with session.chdir("examples/parametric"):
        session.run("bygg")
        session.run("bygg", "--clean")

    with session.chdir("examples/taskrunner"):
        session.run("bygg")
        session.run("bygg", "--clean")

    with session.chdir("examples/trivial"):
        session.run("bygg")
        session.run("bygg", "--clean")

    with session.chdir("examples/environments"):
        session.run("bygg")
        session.run("bygg", "clean_environments")

        session.run("bygg", "default_action")
        session.run("bygg", "clean_environments")

        session.run("bygg", "action1")
        session.run("bygg", "clean_environments")

        session.run("bygg", "action2")
        session.run("bygg", "clean_environments")

        session.run("bygg", "action1", "default_action", "action2")
        session.run("bygg", "clean_environments")



from pathlib import Path
import shutil
import subprocess
import sys
import time
from typing import List

import rich
import rich.status

from bygg.action import Action
from bygg.configuration import (
    PYTHON_INPUTFILE,
    ByggFile,
    Environment,
    load_python_build_file,
)
from bygg.digest import calculate_string_digest
from bygg.scheduler import scheduler
from bygg.util import create_shell_command


def calculate_environment_hash(environment: Environment) -> str:
    requirements: List[str] = []
    requirements.append(environment.shell)
    if environment.inputs:
        for input in environment.inputs:
            with open(input, "r") as f:
                requirements += f.readlines()
    return calculate_string_digest(" ".join(requirements))


loading_python_build_file = rich.status.Status(
    "[cyan]Executing Python build file", spinner="dots"
)


def setup_environment(environment: Environment):
    venv_path = Path(
        environment.venv_directory if environment.venv_directory else ".venv"
    )

    environment_hash = calculate_environment_hash(environment)
    environment_hash_file = venv_path / "bygg_environment_hash.txt"
    if environment_hash_file.exists():
        with open(environment_hash_file, "r") as f:
            if f.read() == environment_hash:
                return True
        environment_hash_file.unlink()

    # TODO should we remove the whole venv or just trust pip to do the right thing? For
    # now, remove the venv.

    if venv_path.exists():
        rich.print(f"[blue]Replacing venv at {venv_path}[/blue]")
        shutil.rmtree(venv_path)

    rich.print("[blue]Setting up environment[/blue]")
    process = subprocess.run(
        environment.shell,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding="utf-8",
    )

    print(process.stdout)

    if process.returncode != 0:
        rich.print(
            f"[red bold]Error while creating virtual environment: [/red bold]{process.stdout}"
        )
        sys.exit(1)

    with open(environment_hash_file, "w") as f:
        f.write(environment_hash)

    return True


def should_restart_with(environment: Environment) -> str | None:
    venv_bin_path = Path(environment.venv_directory) / "bin"

    # # TODO: we could check here if the new venv is using the same Python binary that
    # # we're already running under, and if so just activate the venv.

    is_relative = Path(sys.executable).is_relative_to(venv_bin_path.resolve())
    if not is_relative:
        return str((venv_bin_path / "bygg"))
    return None


def apply_configuration(
    configuration: ByggFile | None,
    environment_name: str | None,
    is_restarted_with_env: str | None,
) -> str | None:
    """Returns the path of the bygg install to restart with if a restart is needed."""
    if configuration:
        if not is_restarted_with_env and environment_name:
            if environment_name in configuration.environments:
                environment = configuration.environments[environment_name]
                setup_environment(environment)

                restart_with = should_restart_with(environment)
                if restart_with is not None:
                    return restart_with

        for action in configuration.actions:
            # TODO Hack until we can iterate over the dependency graph in scheduler.prepare_run:
            if is_restarted_with_env and action.environment != is_restarted_with_env:
                continue

            shell_command = (
                create_shell_command(action.shell, action.message)
                if action.shell
                else None
            )
            Action(
                action.name,
                is_entrypoint=bool(action.is_entrypoint),
                inputs=action.inputs,
                outputs=action.outputs,
                dependencies=action.dependencies,
                command=shell_command,
            )

    # Evaluate the Python build file:

    python_build_file = PYTHON_INPUTFILE
    if configuration and environment_name:
        environment = configuration.environments.get(environment_name, None)
        if environment:
            python_build_file = environment.byggfile

    t0 = time.time()
    action_count = len(scheduler.build_actions)

    with loading_python_build_file:
        load_python_build_file(python_build_file)
    rich.print(
        f"{len(scheduler.build_actions) - action_count} actions registered in "
        f"{time.time() - t0:.2f} seconds."
    )

    return None



from bygg.action import Action, ActionContext
from bygg.common_types import CommandStatus


def hello(ctx: ActionContext):
    print("Hello")
    return CommandStatus(0, "And goodbye.", None)


Action("hello", command=hello, is_entrypoint=True)
